!---------------------------------------------------------------------!
!                                PHAML                                !
!                                                                     !
! The Parallel Hierarchical Adaptive MultiLevel code for solving      !
! linear elliptic partial differential equations of the form          !
! (PUx)x + (QUy)y + RU = F on 2D polygonal domains with mixed         !
! boundary conditions, and eigenvalue problems where F is lambda*U.   !
!                                                                     !
! PHAML is public domain software.  It was produced as part of work   !
! done by the U.S. Government, and is not subject to copyright in     !
! the United States.                                                  !
!                                                                     !
!     William F. Mitchell                                             !
!     Applied and Computational Mathematics Division                  !
!     National Institute of Standards and Technology                  !
!     william.mitchell@nist.gov                                       !
!     http://math.nist.gov/phaml                                      !
!                                                                     !
!---------------------------------------------------------------------!

module message_passing

!----------------------------------------------------
! This module is a dummy version of the message passing library
! to satisfy references in a sequential compilation
!----------------------------------------------------

!----------------------------------------------------
! Other modules used:

use global
use stopwatch
use petsc_init_mod
use slepc_init_mod
use sysdep

implicit none
private
public proc_info, init_comm, terminate_comm, increase_universe, &
       decrease_universe, phaml_send, phaml_recv, phaml_alltoall, &
       phaml_neighbor_alltoall, phaml_global_max, phaml_global_min, &
       phaml_global_sum, PARALLEL, phaml_barrier, phaml_barrier_master, &
       pause_until_enter, warning, fatal, my_proc, num_proc, &
       hostname, graphics_proc, this_processors_procs, pack_procs_size, &
       pack_procs, unpack_procs, cleanup_unpack_procs, slaves_communicator, &
       set_neighbor_procs, get_neighbor_procs, sequential_send, &
       sequential_recv, GRAPHICS_TERMINATE, GRAPHICS_INIT, GRAPHICS_GRID, &
       UNKNOWN_SENDER, NORMAL_SPAWN, DEBUG_SLAVE, DEBUG_GRAPHICS, DEBUG_BOTH, &
       VTUNE_SLAVE, VTUNE_GRAPHICS, VTUNE_BOTH, HOSTLEN

!----------------------------------------------------
! The following types are defined:

integer, parameter :: HOSTLEN = 64
type proc_info
   private
   integer :: nothing ! gotta have something in there, I guess
   logical :: do_graphics
   integer :: graphics_proc
end type proc_info

type message_type
   integer, pointer :: imess(:)
   real(my_real), pointer :: rmess(:)
   integer :: ni, nr
end type message_type

type message_list
   type(message_type), pointer :: message(:)
end type message_list

!----------------------------------------------------
! The following parameters are defined:

integer, parameter :: PARALLEL=SEQUENTIAL
integer, parameter :: UNKNOWN_SENDER = -2
integer, parameter :: GRAPHICS_TERMINATE = 1, &
                      GRAPHICS_INIT      = 2, &
                      GRAPHICS_GRID      = 3
integer, parameter :: NORMAL_SPAWN   = 1, &
                      DEBUG_SLAVE    = 2, &
                      DEBUG_GRAPHICS = 3, &
                      DEBUG_BOTH     = 4, &
                      VTUNE_SLAVE    = 5, &
                      VTUNE_GRAPHICS = 6, &
                      VTUNE_BOTH     = 7
!----------------------------------------------------
! The following variables are defined:

type(proc_info), pointer :: this_processors_procs
!----------------------------------------------------
! The following generic interfaces are defined:
 
interface phaml_global_max
   module procedure phaml_global_max_real, phaml_global_max_int, &
                    phaml_global_max_intv
end interface
 
interface phaml_global_min
   module procedure phaml_global_min_real, phaml_global_min_int
end interface
 
interface phaml_global_sum
   module procedure phaml_global_sum_real, phaml_global_sum_int, &
                    phaml_global_sum_realv,phaml_global_sum_intv
end interface

interface phaml_alltoall
   module procedure phaml_alltoall_int, phaml_alltoall_real, &
                    phaml_alltoall_intv, phaml_alltoall_realv
end interface

interface phaml_neighbor_alltoall
   module procedure phaml_neighbor_alltoall_int, phaml_neighbor_alltoall_real, &
                    phaml_neighbor_alltoall_intv, phaml_neighbor_alltoall_realv
end interface
!----------------------------------------------------

contains

subroutine init_comm(procs,spawn_form,i_draw_grid,master_draws_grid, &
                     graphics_host,output_unit,error_unit,system_size,eq_type, &
                     pde_id,nproc,nthread,max_blen,triangle_files,update_umod, &
                     nalloc,debug_command,display,i_am_graphics)
type (proc_info), target :: procs
integer, intent(inout) :: spawn_form
logical, intent(inout) :: i_draw_grid, master_draws_grid
character(len=*), intent(inout) :: graphics_host
integer, intent(inout) :: output_unit, error_unit, pde_id, system_size, eq_type
integer, intent(in) :: nproc, nthread
real(my_real), intent(inout) :: max_blen
character(len=*), intent(inout) :: triangle_files
logical, intent(inout) :: update_umod
integer, intent(inout) :: nalloc
character(len=*), intent(in), optional :: debug_command, display
logical, intent(in), optional :: i_am_graphics
procs%nothing = 0
procs%do_graphics = i_draw_grid .or. master_draws_grid
! set graphics_proc -1 if not doing graphics, and 1 if doing graphics
if (procs%do_graphics) then
   procs%graphics_proc = 1
else
   procs%graphics_proc = -1
endif
! set graphics_proc to -2 if I am the graphics proc, -1 if not
! doing graphics, and 1 if I am not the graphics proc and doing graphics
if (present(i_am_graphics)) then
   procs%graphics_proc = -2
elseif (procs%do_graphics) then
   procs%graphics_proc = 1
else
   procs%graphics_proc = -1
endif
this_processors_procs => procs
update_umod = .false.
call slepc_init(0)
call petsc_init(0)
end subroutine init_comm

!          ---------------
subroutine sequential_send(imess,ni,rmess,nr)
!          ---------------

!----------------------------------------------------
! This routine sends a message to graphics when running a sequential program
!----------------------------------------------------

!----------------------------------------------------
! Dummy arguments

integer, intent(in) :: imess(:),ni,nr
real(my_real), intent(in) :: rmess(:)
!----------------------------------------------------
! Local variables:

logical :: exists, opened, firstcall=.true., secondcall=.false.
integer :: iounit
type(watchtype) :: w
real :: time_read
integer :: nwarn
!----------------------------------------------------
! Begin executable code

! on the first call, if the message files exist, delete them

if (firstcall) then
   iounit = 11
   do
      inquire(unit=iounit,exist=exists,opened=opened)
      if (exists .and. .not. opened) exit
      iounit = iounit + 1
   end do
   inquire(file=trim(scratch)//"phaml_message",exist=exists)
   if (exists) then
      open(iounit,file=trim(scratch)//"phaml_message")
      close(iounit,status="DELETE")
   endif
   inquire(file=trim(scratch)//"phaml_lock",exist=exists)
   if (exists) then
      open(iounit,file=trim(scratch)//"phaml_lock")
      close(iounit,status="DELETE")
   endif
endif

! If the message file exists, wait for the graphics program to delete it.
! On the second call, give periodic warnings if it exists for a long time and
! eventually quit if persisits.

if (secondcall) then
   call create_watch(w)
   call start_watch(w)
   nwarn = 1
endif
do
   inquire(file=trim(scratch)//"phaml_message",exist=exists)
   if (.not. exists) exit
   if (secondcall) then
      call read_watch(time_read,w,"wall")
      if (time_read > 300.0) then
         call fatal("Have been waiting for the graphics process to read the message for 5 minutes.","Quitting.")
         stop
      endif
      if (time_read > nwarn*15.0) then
         nwarn = nwarn + 1
         call warning("This seems to be a sequential run using graphics, without the graphics process.", &
                      "You must run phaml_graphics by hand.")
      endif
   endif

end do 
if (secondcall) then
   call destroy_watch(w)
   secondcall = .false.
endif
if (firstcall) then
   firstcall = .false.
   secondcall = .true.
endif

! find an available io unit

iounit = 11
do
   inquire(unit=iounit,exist=exists,opened=opened)
   if (exists .and. .not. opened) exit
   iounit = iounit + 1
end do

! create a lock file to indicate data is being written

open(unit=iounit,file=trim(scratch)//"phaml_lock")
write(iounit,*) 0
close(iounit)

! write the message

open(unit=iounit,file=trim(scratch)//"phaml_message")
write(iounit,*) ni, nr
write(iounit,*) imess
write(iounit,*) rmess
close(iounit)

! delete the lock file

open(unit=iounit,file=trim(scratch)//"phaml_lock")
close(unit=iounit,status="DELETE")

end subroutine sequential_send

!          ---------------
subroutine sequential_recv(imess,ni,rmess,nr,no_message)
!          ---------------

!----------------------------------------------------
! This routine receives a message from the program when running sequentially
!----------------------------------------------------

!----------------------------------------------------
! Dummy arguments

integer, pointer :: imess(:)
real(my_real), pointer :: rmess(:)
integer, intent(out) :: ni, nr
logical, intent(out) :: no_message
!----------------------------------------------------
! Local variables:

logical :: exists, opened
integer :: allocstat, iounit
!----------------------------------------------------
! Begin executable code

! flag for no message yet

no_message = .true.
ni = 0
nr = 0

! see if the message file exists

inquire(file=trim(scratch)//"phaml_message",exist=exists)
if (.not. exists) return

! see if the message is complete

inquire(file=trim(scratch)//"phaml_lock",exist=exists)
if (exists) return

! we have a message

no_message = .false.

! find an available io unit

iounit = 11
do
   inquire(unit=iounit,exist=exists,opened=opened)
   if (exists .and. .not. opened) exit
   iounit = iounit + 1
end do

! read the message file

open(unit=iounit,file=trim(scratch)//"phaml_message")
read(iounit,*) ni, nr
allocate(imess(ni),rmess(nr),stat=allocstat)
if (allocstat /= 0) then
   call fatal("allocation failed for receiving message",intlist=(/allocstat/))
   stop
endif
read(iounit,*) imess
read(iounit,*) rmess

! delete message file

close(unit=iounit,status="DELETE")

end subroutine sequential_recv

subroutine terminate_comm(procs,finalize_mpi)
type (proc_info), intent(inout) :: procs
logical, intent(in) :: finalize_mpi
if (procs%do_graphics) then
   call sequential_send((/GRAPHICS_TERMINATE/),1,(/0.0_my_real/),0)
endif
if (finalize_mpi) then
   call slepc_finalize
   call petsc_finalize
endif
end subroutine terminate_comm

subroutine increase_universe(procs,recv_int)
type (proc_info), intent(inout) :: procs
integer, intent(in) :: recv_int(:)
end subroutine increase_universe

subroutine decrease_universe(procs)
type (proc_info), intent(inout) :: procs
end subroutine decrease_universe

subroutine phaml_send(procs,proc,int_message,ni,real_message,nr,tag)
type (proc_info), intent(in) :: procs
integer, intent(in) :: proc, tag,ni,nr
integer, intent(in) :: int_message(ni)
real(my_real), intent(in) :: real_message(nr)
end subroutine phaml_send

subroutine phaml_recv(procs,proc,int_message,ni,real_message,nr,tag,noblock)
type (proc_info), intent(in) :: procs
integer, intent(in) :: tag
integer, intent(out) :: proc,ni,nr
integer, pointer :: int_message(:)
real(my_real), pointer :: real_message(:)
logical, optional, intent(in) :: noblock
proc=1
ni=0
nr=0
nullify(int_message,real_message)
end subroutine phaml_recv

subroutine phaml_alltoall_int(procs,send,nsend,recv,nrecv,tag)
type (proc_info), intent(in) :: procs
integer, intent(in) :: send(:)
integer, intent(in) :: nsend
integer, pointer :: recv(:)
integer, intent(out) :: nrecv(:)
integer, intent(in) :: tag
end subroutine phaml_alltoall_int

subroutine phaml_alltoall_real(procs,send,nsend,recv,nrecv,tag)
type (proc_info), intent(in) :: procs
real(my_real), intent(in) :: send(:)
integer, intent(in) :: nsend
real(my_real), pointer :: recv(:)
integer, intent(out) :: nrecv(:)
integer, intent(in) :: tag
end subroutine phaml_alltoall_real

subroutine phaml_alltoall_intv(procs,send,nsend,recv,nrecv,tag)
type (proc_info), intent(in) :: procs
integer, intent(in) :: send(:)
integer, intent(in) :: nsend(:)
integer, pointer :: recv(:)
integer, intent(out) :: nrecv(:)
integer, intent(in) :: tag
end subroutine phaml_alltoall_intv

subroutine phaml_alltoall_realv(procs,send,nsend,recv,nrecv,tag)
type (proc_info), intent(in) :: procs
real(my_real), intent(in) :: send(:)
integer, intent(in) :: nsend(:)
real(my_real), pointer :: recv(:)
integer, intent(out) :: nrecv(:)
integer, intent(in) :: tag
end subroutine phaml_alltoall_realv

subroutine phaml_neighbor_alltoall_int(procs,send,nsend,recv,nrecv,tag)
type (proc_info), intent(in) :: procs
integer, intent(in) :: send(:)
integer, intent(in) :: nsend
integer, pointer :: recv(:)
integer, intent(out) :: nrecv(:)
integer, intent(in) :: tag
end subroutine phaml_neighbor_alltoall_int

subroutine phaml_neighbor_alltoall_real(procs,send,nsend,recv,nrecv,tag)
type (proc_info), intent(in) :: procs
real(my_real), intent(in) :: send(:)
integer, intent(in) :: nsend
real(my_real), pointer :: recv(:)
integer, intent(out) :: nrecv(:)
integer, intent(in) :: tag
end subroutine phaml_neighbor_alltoall_real

subroutine phaml_neighbor_alltoall_intv(procs,send,nsend,recv,nrecv,tag)
type (proc_info), intent(in) :: procs
integer, intent(in) :: send(:)
integer, intent(in) :: nsend(:)
integer, pointer :: recv(:)
integer, intent(out) :: nrecv(:)
integer, intent(in) :: tag
end subroutine phaml_neighbor_alltoall_intv

subroutine phaml_neighbor_alltoall_realv(procs,send,nsend,recv,nrecv,tag)
type (proc_info), intent(in) :: procs
real(my_real), intent(in) :: send(:)
integer, intent(in) :: nsend(:)
real(my_real), pointer :: recv(:)
integer, intent(out) :: nrecv(:)
integer, intent(in) :: tag
end subroutine phaml_neighbor_alltoall_realv

function phaml_global_max_real(procs,var,tag,just1)
type (proc_info), intent(in) :: procs
real (my_real) phaml_global_max_real
real (my_real), intent(in) :: var
integer, intent(in) :: tag
logical, optional, intent(in) :: just1
phaml_global_max_real = var
end function phaml_global_max_real

function phaml_global_max_int(procs,var,tag,just1)
type (proc_info), intent(in) :: procs
integer phaml_global_max_int
integer, intent(in) :: var
integer, intent(in) :: tag
logical, optional, intent(in) :: just1
phaml_global_max_int = var
end function phaml_global_max_int

function phaml_global_max_intv(procs,var,tag,just1)
type (proc_info), intent(in) :: procs
integer, intent(in) :: var(:)
integer, intent(in) :: tag
logical, optional, intent(in) :: just1
integer, dimension(size(var)) :: phaml_global_max_intv
phaml_global_max_intv = var
end function phaml_global_max_intv

function phaml_global_min_real(procs,var,tag,just1)
type (proc_info), intent(in) :: procs
real (my_real) phaml_global_min_real
real (my_real), intent(in) :: var
integer, intent(in) :: tag
logical, optional, intent(in) :: just1
phaml_global_min_real = var
end function phaml_global_min_real

function phaml_global_min_int(procs,var,tag,just1)
type (proc_info), intent(in) :: procs
integer phaml_global_min_int
integer, intent(in) :: var
integer, intent(in) :: tag
logical, optional, intent(in) :: just1
phaml_global_min_int = var
end function phaml_global_min_int

function phaml_global_sum_real(procs,var,tag,just1)
type (proc_info), intent(in) :: procs
real (my_real) phaml_global_sum_real
real (my_real), intent(in) :: var
integer, intent(in) :: tag
logical, optional, intent(in) :: just1
phaml_global_sum_real = var
end function phaml_global_sum_real

function phaml_global_sum_int(procs,var,tag,just1)
type (proc_info), intent(in) :: procs
integer phaml_global_sum_int
integer, intent(in) :: var
integer, intent(in) :: tag
logical, optional, intent(in) :: just1
phaml_global_sum_int = var
end function phaml_global_sum_int

function phaml_global_sum_realv(procs,var,tag,just1)
type (proc_info), intent(in) :: procs
real(my_real), intent(in) :: var(:)
integer, intent(in) :: tag
logical, optional, intent(in) :: just1
real, dimension(size(var)) :: phaml_global_sum_realv
phaml_global_sum_realv = var
end function phaml_global_sum_realv

function phaml_global_sum_intv(procs,var,tag,just1)
type (proc_info), intent(in) :: procs
integer, intent(in) :: var(:)
integer, intent(in) :: tag
logical, optional, intent(in) :: just1
integer, dimension(size(var)) :: phaml_global_sum_intv
phaml_global_sum_intv = var
end function phaml_global_sum_intv

function my_proc(procs)
integer :: my_proc
type (proc_info), intent(in) :: procs
my_proc = 1
end function my_proc

function num_proc(procs)
integer :: num_proc
type (proc_info), intent(in) :: procs
num_proc = 1
end function num_proc

function hostname(procs)
character(len=HOSTLEN) hostname
type (proc_info), intent(in) :: procs
hostname = " "
end function hostname

function graphics_proc(procs)
integer :: graphics_proc
type (proc_info), intent(in) :: procs
graphics_proc = procs%graphics_proc
end function graphics_proc

function slaves_communicator(procs)
integer :: slaves_communicator ! note default integer
type (proc_info), intent(in) :: procs
slaves_communicator = 0
end function slaves_communicator

subroutine pack_procs_size(procs,ni,nr)
type(proc_info), intent(in) :: procs
integer, intent(out) :: ni,nr
ni = 0
nr = 0
end subroutine pack_procs_size

subroutine pack_procs(procs,send_int,first_int,send_real,first_real)
type(proc_info), intent(in) :: procs
integer, intent(inout) :: send_int(:)
real(my_real), intent(inout) :: send_real(:)
integer, intent(inout) :: first_int, first_real
end subroutine pack_procs

subroutine unpack_procs(procs,send_int,first_int,send_real,first_real)
type(proc_info), intent(out) :: procs
integer, intent(in) :: send_int(:)
real(my_real), intent(in) :: send_real(:)
integer, intent(inout) :: first_int, first_real
end subroutine unpack_procs

subroutine cleanup_unpack_procs(procs)
type(proc_info), intent(inout) :: procs
end subroutine cleanup_unpack_procs

subroutine set_neighbor_procs(procs,neighbors)
type(proc_info), intent(inout) :: procs
integer, intent(in) :: neighbors(:)
end subroutine set_neighbor_procs

subroutine get_neighbor_procs(procs,neighbors,self)
type(proc_info), intent(in) :: procs
integer, pointer, intent(out) :: neighbors(:)
logical, intent(in), optional :: self
nullify(neighbors)
end subroutine get_neighbor_procs

subroutine phaml_barrier(procs)
type(proc_info), intent(in) :: procs
end subroutine phaml_barrier

subroutine phaml_barrier_master(procs)
type(proc_info), intent(in) :: procs
end subroutine phaml_barrier_master

!          -----------------
subroutine pause_until_enter(procs,really_pause,just_one,dont_pausewatch)
!          -----------------
 
!----------------------------------------------------
! This routine pauses until the return (or enter) key is pressed.  The
! master and all slaves must call this routine to avoid deadlock.
! If really_pause is false, the pause does not occur.
! If just_one is present and true, the 'pause complete' message is sent
! from the master to only slave 1.
!----------------------------------------------------
 
!----------------------------------------------------
! Dummy arguments
 
type(proc_info), intent(in) :: procs
logical, intent(in) :: really_pause
logical, intent(in), optional :: just_one, dont_pausewatch
 
!----------------------------------------------------
! Local variables:

!----------------------------------------------------
! Begin executable code

if (really_pause) then
   if (present(dont_pausewatch)) then
      if (.not. dont_pausewatch) call pause_watch(all_watches)
   else
      call pause_watch(all_watches)
   endif
   write(outunit,"(A)") "press return to continue"
   call my_flush(outunit)
   read *
   if (present(dont_pausewatch)) then
      if (.not. dont_pausewatch) call end_pause_watch(all_watches)
   else
      call end_pause_watch(all_watches)
   endif
endif

end subroutine pause_until_enter

!          -------
subroutine warning(msg,msg2,intlist,reallist)
!          -------

!----------------------------------------------------
! This routine handles warning messages
!----------------------------------------------------

!----------------------------------------------------
! Dummy arguments

character(len=*), intent(in) :: msg
character(len=*), intent(in), optional :: msg2
integer, intent(in), optional, dimension(:) :: intlist
real(my_real), intent(in), optional, dimension(:) :: reallist
!----------------------------------------------------
! Local variables:

!----------------------------------------------------
! Begin executable code

if (warn) then
   write(errunit,"(A)")
   write(errunit,"(A)") "------------------------------------------------------"
   write(errunit,"(3A)") "          PHAML Version ",version_number," WARNING"
   write(errunit,"(A)") msg
   if (present(msg2)) write(errunit,"(A)") msg2
   if (present(intlist)) write(errunit,"(7I11)") intlist
   if (present(reallist)) write(errunit,"(SS,1P,4E18.10E2)") reallist
   write(errunit,"(A)") "------------------------------------------------------"
   write(errunit,"(A)")
endif

end subroutine warning

!          -----
subroutine fatal(msg,msg2,intlist,reallist,procs)
!          -----

!----------------------------------------------------
! This routine handles fatal errors
!----------------------------------------------------

!----------------------------------------------------
! Dummy arguments

character(len=*), intent(in) :: msg
character(len=*), intent(in), optional :: msg2
integer, intent(in), optional, dimension(:) :: intlist
real(my_real), intent(in), optional, dimension(:) :: reallist
type(proc_info), intent(in), optional :: procs
!----------------------------------------------------
! Local variables:

!----------------------------------------------------
! Begin executable code

write(errunit,"(A)")
write(errunit,"(A)") "------------------------------------------------------"
write(errunit,"(3A)") "          PHAML Version ",version_number," ERROR"
write(errunit,"(A)") msg
if (present(msg2)) write(errunit,"(A)") msg2
if (present(intlist)) write(errunit,"(7I11)") intlist
if (present(reallist)) write(errunit,"(SS,1P,4E18.10E2)") reallist
write(errunit,"(A)") "------------------------------------------------------"
write(errunit,"(A)")

stop

end subroutine fatal

end module message_passing
